
/**
 *  All functions below are modified from native Fx Customize code in
 *  ../toolkit/customizeToolbar.js, as of changeset:
 *
 *  40690e93dfc0 2009-06-22 20:29 +0800.
 *
 *  The major changes are to wrap functions in a for(){} loop to handle multiple
 *  toolboxes, to change the reference from the single gToolbox to multiple
 *  toolbox elements within the loop, support for statusbar as a toolbar.
 */

InitWithToolbox = function (aToolbox) {
  // Get the navToolbox (original toolbox, containing defaults) and statusbar.
  var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"]
                              .getService(Components.interfaces.nsIWindowMediator);
  var win = wm.getMostRecentWindow("navigator:browser");
//  if (win) {
    gStatusbar = win.document.getElementById("status-bar");
    navToolbox = win.gNavToolbox; //getNavToolbox();
//  }

  if (gToolboxSheet) {
    // Note: if gToolboxSheet (customize dialog in an iframe) then aToolbox is
    // our array, as there won't be a window; make gToolbox be the navToolbox.
    gToolbox = navToolbox;
    gToolboxes = aToolbox;
    winDoc = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                   .getInterface(Components.interfaces.nsIWebNavigation)
                   .QueryInterface(Components.interfaces.nsIDocShellTreeItem)
                   .rootTreeItem
                   .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                   .getInterface(Components.interfaces.nsIDOMWindow)
                   .document;
  }
  else {
    // Pass the full array of toolboxes as the 2nd argument so as not to
    // break any other xul app (extension) using customisable toolbars.
    // Note: statusbar is both a 'toolbox' and a 'toolbar' here..
    gToolbox = aToolbox;
    gToolboxes = (window.arguments && window.arguments.length > 1) ?
                  window.arguments[1] : [aToolbox];
    winDoc = window.opener.top.document;
  }

  // If statusbar, it must be the 0 (and only) index in the array.
  gStatusbarMode = gToolboxes[0].localName == "statusbar" ? true : false;

  gToolboxDocument = gToolbox.ownerDocument;

  for(var i = 0; i < gToolboxes.length; i++) {
    gToolboxes[i].addEventListener("dragstart", onToolbarDragStart, false);
    gToolboxes[i].addEventListener("dragover", onToolbarDragOver, false);
    gToolboxes[i].addEventListener("dragleave", onToolbarDragLeave, false);
    gToolboxes[i].addEventListener("drop", onToolbarDrop, false);
    // For css styling of min-width etc.
    gToolboxes[i].customizing = true;
  }

  if (gStatusbarMode) {
    gToolbox = gStatusbar;
    var dd = gStatusbar.getAttribute("ondrop");
    if (dd) {
      gStatusbar.setAttribute("tt_ondrop", dd);
      gStatusbar.removeAttribute("ondrop");
    }
  }

  initDialog();
}

function initDialog() {
  // Move box into groupbox..
  var globalBox = document.getElementById("smallicons").parentNode;
  document.getElementById("tt-globalGroup").appendChild(globalBox);
  document.getElementById("smallicons").setAttribute("flex", 1);

  if (!gStatusbarMode) {
    var mode = navToolbox.getAttribute("mode");
    document.getElementById("modelist").value = mode;
    gToolboxIconSize = navToolbox.getAttribute("iconsize");
    var smallIconsCheckbox = document.getElementById("smallicons");
    smallIconsCheckbox.checked = gToolboxIconSize == "small";
    if (mode == "text")
      smallIconsCheckbox.disabled = true;

    // Hide global groupbox description for other apps (Console2 eg)
    var winType = winDoc.documentElement.getAttribute("windowtype");
    if (winType == "navigator:browser")
      document.getElementById("tt-globalDesc").hidden = false;
    else
      document.getElementById("tt-globalDesc").hidden = true;

    // Unset fixed width for vertical toolboxes while in Customize
    if (window.opener && window.opener.top &&
        "tt_adjustToolboxWidth" in window.opener.top)
      // Customize window.
      window.opener.top.tt_adjustToolboxWidth(false);
    else
      // Customize sheet.
      tt_adjustToolboxWidth(false);
  }
  else {
    var stringBundle = document.getElementById("totaltoolbarBundle");
    var title = stringBundle.getString("dialog.title.statusbar");
    document.title = title;
    var instructions = stringBundle.getString("instructions.description.statusbar");
    document.getElementById("instructions").setAttribute("value", instructions);
    document.getElementById("tt-globalGroupDeck").setAttribute("selectedIndex", 1);
  }

  // Build up the palette of other items.
  buildPalette();

  // Wrap all the items on the toolbar in toolbarpaletteitems.
  wrapToolbarItems();
}

repositionDialog = function () {
  // Always use persisted dimensions and position!
  return;
}

removeToolboxListeners = function () {
  for(var i = 0; i < gToolboxes.length; i++) {
    gToolboxes[i].removeEventListener("dragstart", onToolbarDragStart, false);
    gToolboxes[i].removeEventListener("dragover", onToolbarDragOver, false);
    gToolboxes[i].removeEventListener("dragleave", onToolbarDragLeave, false);
    gToolboxes[i].removeEventListener("drop", onToolbarDrop, false);
    gToolboxes[i].customizing = false;
  }
}

/**
 * Persist the current set of buttons in all customizable toolbars to
 * localstore.
 */
//TODO: make reusable function
function persistCurrentSets() {
  if (!gToolboxChanged || gToolboxDocument.defaultView.closed)
    return;

  docElt = gToolboxDocument.documentElement;
  
  for(var j = 0; j < gToolboxes.length; j++) {
    var toolbox = gToolboxes[j];
    var len = gStatusbarMode ? gToolboxes.length : gToolboxes[j].childNodes.length;

    var customCount = 0;
    var toolboxEmpty = true;
    for(var i = 0; i < len; ++i) {
      // Look for customizable toolbars that need to be persisted.  If toolbar
      // added but no buttons, arrays will not have it, so handle |i| node
      // undefined.
      var toolbar = gStatusbarMode ? toolbox :
                    toolbox.childElementCount > 0 ? (toolbox.childNodes[i] ?
                    toolbox.childNodes[i] : null) : null;
      if (isCustomizableToolbar(toolbar)) {
        // Calculate currentset and store it in the attribute.
        var currentSet = toolbar.currentSet;
        toolbar.setAttribute("currentset", currentSet);

        var customIndex = toolbar.hasAttribute("customindex");
        if (customIndex) {
          if (!toolbar.firstChild) {
            // Remove any options from root doc and
            handleOptions("remove", toolbar);
            // Remove custom toolbars whose contents have been removed.
            toolbox.removeChild(toolbar);
            // Bump count since an item is gone.
            i++;
          } 
          else {
            toolboxEmpty = false;
            if (toolbox.toolbarset.hasAttribute('anonymous')) {
              if (currentSet != "__empty") {
                // Must persist xbl 'below-tabs' toolbar on the document root
                var attrPrefix = '_toolbarset.' +
                    toolbox.toolbarset.getAttribute('id') + '.toolbar';
                docElt.setAttribute(attrPrefix +
                    (++customCount), toolbar.toolbarName + ":" + currentSet);
                toolbar.setAttribute("customindex", customCount); //adjust count
                gToolboxDocument.persist(docElt.id, attrPrefix+customCount);
              }
            } 
            else {
              // Persist custom toolbar info on the <toolbarset/>
              toolbox.toolbarset.setAttribute("toolbar" + (++customCount),
                  toolbar.toolbarName + ":" + currentSet);
              toolbar.setAttribute("customindex", customCount); //adjust count
              gToolboxDocument.persist(toolbox.toolbarset.id,
                  "toolbar" + customCount);
            }
          }
        }

        if (!customIndex) {
          if (!toolbar.firstChild)
            // Remove any options from root doc
            handleOptions("remove", toolbar);
          else
            toolboxEmpty = false;
          if (currentSet != "__empty" || toolbar.hasAttribute("defaultset")) {
            if (toolbar.hasAttribute('anonymous')) {
              // Must persist xbl tabbar toolbars on the document root
              var attr = "_toolbar."+toolbar.getAttribute('id')+".currentset";
              docElt.setAttribute(attr, currentSet);
              gToolboxDocument.persist(gToolboxDocument.documentElement.id, attr);
            }
          }
          else
            if (!toolbar.hasAttribute("defaultset"))
              toolbar.removeAttribute("currentset");
        // Persist the currentset attribute directly on hardcoded toolbars.
        gToolboxDocument.persist(toolbar.id, "currentset");
        }
      }
    }
    // Number of remaining toolbars in toolbox
    toolbox.customToolbarCount = customCount;
    setAttribute(toolbox, "empty", toolboxEmpty);
    gToolboxDocument.persist(toolbox.id, "empty");

    // Remove toolbarX attributes from toolbarset for removed toolbars, ie any
    // entries past those for toolbars already indexed..
    if (toolbox.toolbarset) {
      if (toolbox.toolbarset.hasAttribute('anonymous')) {
        var attrPrefix = '_toolbarset.' +
            toolbox.toolbarset.getAttribute('id') + '.toolbar';
        while (docElt.hasAttribute(attrPrefix+(++customCount))) {
          docElt.removeAttribute(attrPrefix+customCount);
          gToolboxDocument.persist(docElt.id, attrPrefix+customCount);
        }
      }
      else {
        while (toolbox.toolbarset.hasAttribute("toolbar"+(++customCount))) {
          toolbox.toolbarset.removeAttribute("toolbar"+customCount);
          gToolboxDocument.persist(toolbox.toolbarset.id, "toolbar"+customCount);
        }
      }
    }
  }
}

/**
 * Wraps all items in all customizable toolbars in a toolbox.
 */
wrapToolbarItems = function () {
  if (gStatusbarMode)
    winDoc.getElementById("tt-statusbar-spacer").setAttribute("flex", 1);

  for(var j = 0; j < gToolboxes.length; j++) {
    var toolbox = gStatusbarMode ? gToolboxes : gToolboxes[j];
    var len = gStatusbarMode ? gToolboxes.length : gToolboxes[j].childNodes.length;

    for(var i = 0; i < len; ++i) {
      var toolbar = (gStatusbarMode) ?
              toolbox[i] : (toolbox.childNodes ? toolbox.childNodes[i] : null);
      if (isCustomizableToolbar(toolbar)) {
        for(var k = 0; k < toolbar.childNodes.length; ++k) {
          var item = toolbar.childNodes[k];

          // Bug 317062 fix for MacOS
          if (item.firstChild && item.firstChild.localName == "menubar")
            continue;

          if (isToolbarItem(item)) {
            var wrapper = wrapToolbarItem(item);
            cleanupItemForToolbar(item, wrapper);
          }
        }
      }
    }
  }
}

/**
 * Unwraps all items in all customizable toolbars in a toolbox.
 */
unwrapToolbarItems = function () {
  for(var j = 0; j < gToolboxes.length; j++) {
    var toolbox = gToolboxes[j];

    var paletteItems = toolbox.getElementsByTagName("toolbarpaletteitem");
    var paletteItem;
    while ((paletteItem = paletteItems.item(0)) != null) {
      var toolbarItem = paletteItem.firstChild;

      if (paletteItem.hasAttribute("itemdisabled"))
        toolbarItem.disabled = true;

      if (paletteItem.hasAttribute("itemcommand"))
        toolbarItem.setAttribute("command", paletteItem.getAttribute("itemcommand"));

      if (paletteItem.hasAttribute("itemhidden"))
        toolbarItem.hidden = true;

      if (paletteItem.hasAttribute("itemcollapsed"))
        toolbarItem.collapsed = true;

      // Remove label child from statusbarpanel item, if added..
      if (paletteItem.hasAttribute("itemlabelled")) {
        if (toolbarItem.className == "statusbarpanel-iconic-text")
          toolbarItem.removeAttribute("label");
        else
          toolbarItem.removeChild(toolbarItem.lastChild);
      }

      // Restore style if any, remove forced display
      toolbarItem.removeAttribute("style");
      if (paletteItem.hasAttribute("itemstyle"))
        toolbarItem.setAttribute("style", paletteItem.getAttribute("itemstyle"));

      paletteItem.parentNode.replaceChild(toolbarItem, paletteItem);
    }
  }

  if (gStatusbarMode)
    winDoc.getElementById("tt-statusbar-spacer").setAttribute("flex", 1);
}

/**
 * Wraps an item that is currently on a toolbar and replaces the item
 * with the wrapper. This is not used when dropping items from the palette,
 * only when first starting the dialog and wrapping everything on the toolbars.
 */
function wrapToolbarItem(aToolbarItem) {
  var wrapper = createWrapper(aToolbarItem.id, gToolboxDocument);

  wrapper.flex = aToolbarItem.localName == "toolbarspring" ? 1 : 0;

  aToolbarItem.parentNode.replaceChild(wrapper, aToolbarItem);

  wrapper.appendChild(aToolbarItem);

  return wrapper;
}

/**
 * Get the list of ids for the current set of items on each toolbar.
 */
getCurrentItemIds = function () {
  var currentItems = {};
  for(var j = 0; j < gToolboxes.length; j++) {
    var toolbox = (gStatusbarMode) ? gToolboxes : gToolboxes[j];
    var len = gStatusbarMode ? gToolboxes.length : gToolboxes[j].childNodes.length;

    for (var i = 0; i < len; ++i) {
      var toolbar = (gStatusbarMode) ?
          toolbox[i] : (toolbox.childNodes ? toolbox.childNodes[i] : null);
      if (isCustomizableToolbar(toolbar)) {
        var child = toolbar.firstChild;
        while (child) {
          if (isToolbarItem(child) && !child.getAttribute("removed"))
            currentItems[child.id] = 1;
          child = child.nextSibling;
        }
      }
    }
  }

  return currentItems;
}

/**
 * Makes sure that an item that has been cloned from a template
 * is stripped of all properties that may adversely affect its
 * appearance in the palette.
 */
function cleanUpItemForPalette(aItem, aWrapper) {
  aWrapper.setAttribute("place", "palette");
  setWrapperType(aItem, aWrapper);

  if (aItem.hasAttribute("title"))
    aWrapper.setAttribute("title", aItem.getAttribute("title"));
  else if (isSpecialItem(aItem)) {
      var stringBundle = document.getElementById("stringBundle");
      // Remove the common "toolbar" prefix to generate the string name.
      var title = stringBundle.getString(aItem.localName.slice(7) + "Title");
      aWrapper.setAttribute("title", title);
  }
  else if (!aItem.hasAttribute("label")) {
    var title = aItem.hasAttribute("tooltiptext") ?
        aItem.getAttribute("tooltiptext") : aItem.id;
    aWrapper.setAttribute("title", title);
  }

  // Remove attributes that screw up our appearance.
  aItem.removeAttribute("command");
  aItem.removeAttribute("observes");
  aItem.removeAttribute("disabled");
  aItem.removeAttribute("type");
  aItem.removeAttribute("width");
  aItem.removeAttribute("hidden");
  aItem.removeAttribute("collapsed");

  if (aItem.localName == "toolbaritem" && aItem.firstChild) {
    aItem.firstChild.removeAttribute("observes");

    // So the throbber doesn't throb in the dialog,
    // cute as that may be...
    aItem.firstChild.removeAttribute("busy");
  }
}

/**
 * Makes sure that an item that has been cloned from a template
 * is stripped of all properties that may adversely affect its
 * appearance in the toolbar.  Store critical properties on the
 * wrapper so they can be put back on the item when we're done.
 */
function cleanupItemForToolbar(aItem, aWrapper) {
  setWrapperType(aItem, aWrapper);
  aWrapper.setAttribute("place", "toolbar");

  if (aItem.hasAttribute("command")) {
    aWrapper.setAttribute("itemcommand", aItem.getAttribute("command"));
    aItem.removeAttribute("command");
  }

  if (aItem.disabled) {
    aWrapper.setAttribute("itemdisabled", true);
    aItem.disabled = false;
  }

  // The rest is for statusbar.
  if (!gStatusbarMode)
    return;

  if (aItem.hidden) {
    aWrapper.setAttribute("itemhidden", true);
    aItem.removeAttribute("hidden");
  }

  if (aItem.collapsed) {
    aWrapper.setAttribute("itemcollapsed", true);
    aItem.removeAttribute("collapsed");
  }

  if (aItem.minWidth > 0) {
    aWrapper.setAttribute("itemminwidth", aItem.minWidth);
    aItem.minWidth = 0;
  }
  
  if (aItem.hasAttribute("removed")) {
    aWrapper.setAttribute("itemremoved", aItem.getAttribute("removed"));
  }

  // Add a label to statusbarpanel, for difficult to id statusbar items
  if (aItem.getAttribute("class") == "statusbarpanel-iconic-text" &&
      !aItem.hasAttribute("label"))
    aItem.setAttribute("label", aItem.hasAttribute("title") ?
        aItem.getAttribute("title") : aItem.id);

  // If no width (must allow for border etc incustomizemode), then no label
  // or icon, so create label and insert id to identify the thing..
  if (aItem.boxObject.width <= 4 && !isSpecialItem(aItem) &&
      aItem.getAttribute("class") != "statusbarpanel-iconic-text" &&
      !aItem.hasAttribute("label")) {
    // Add a label to statusbarpanel, for difficult to id statusbar items
//    else 
//      if (!aItem.hasAttribute("label")) {
        var label = document.createElement("label");
        gToolboxDocument.adoptNode(label);
        aItem.appendChild(label);
        aItem.lastChild.setAttribute("value",
            aItem.hasAttribute("label") ? aItem.getAttribute("label") :
            (aItem.hasAttribute("title") ? aItem.getAttribute("title") :
            aItem.id));
        aWrapper.setAttribute("itemlabelled", true);
      }
//  }

  // Force any "display: none" items to show up..  Set after boxObject check.
  if (aItem.hasAttribute("style"))
    aWrapper.setAttribute("itemstyle", aItem.getAttribute("style"));
  if (!aItem.hasAttribute("removed"))
    aItem.setAttribute("style", "display: -moz-box !important");
}

function setWrapperType(aItem, aWrapper) {
  if (aItem.localName == "toolbarseparator") {
    aWrapper.setAttribute("type", "separator");
  } else if (aItem.localName == "toolbarspring") {
    aWrapper.setAttribute("type", "spring");
  } else if (aItem.localName == "toolbarspacer") {
    aWrapper.setAttribute("type", "spacer");
  } else if (aItem.localName == "toolbaritem" && aItem.firstChild) {
    aWrapper.setAttribute("type", aItem.firstChild.localName);
  } else if (aItem.localName == "statusbarpanel") {
    aWrapper.setAttribute("type", aItem.localName);
  }
}

function setDragActive(aItem, aValue) {
  var node = aItem;
  var direction = window.getComputedStyle(aItem, null).direction;
  var value = direction == "ltr"? "left" : "right";
  if (aItem.localName == "toolbar") {
    node = aItem.lastChild || aItem;
    value = direction == "ltr"? "right" : "left";
  }

  if (!node)
    return;

  if (aValue) {
    if (!node.hasAttribute("dragover"))
      node.setAttribute("dragover", value);
    } else {
      node.removeAttribute("dragover");
    }
}

addNewToolbar = function () {
  var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                                .getService(Components.interfaces.nsIPromptService);

  var stringBundle = document.getElementById("stringBundle");
  var message = stringBundle.getString("enterToolbarName");
  var title = stringBundle.getString("enterToolbarTitle");

  var name = {};

  // Quitting from the toolbar dialog while the new toolbar prompt is up
  // can cause things to become unresponsive on the Mac. Until dialog modality
  // is fixed (395465), disable the "Done" button explicitly.
  var doneButton = document.getElementById("donebutton");
  if (doneButton)
    doneButton.disabled = true;

  while(true) {

    if (!promptService.prompt(window, title, message, name, null, {})) {
      if (doneButton)
        doneButton.disabled = false;
      return;
    }

    if (!name.value) {
      message = stringBundle.getFormattedString("enterToolbarBlank", [name.value]);
     continue;
    }
    //TODO: need better error msg string
    if (name.value.match(/[^\s|^\w]/g)) {
       message = stringBundle.getFormattedString("enterToolbarBlank", [name.value]);
       continue;
    }

    var dupeFound = false;

    // Check for an existing toolbar with the same display name
    for(var j = 0; j < gToolboxes.length; j++) {
      var toolbox = gToolboxes[j];

      for(var i = 0; i < toolbox.childNodes.length; ++i) {
        var toolbar = toolbox.childNodes[i];
        var toolbarName = toolbar.getAttribute("toolbarname");

        if (toolbarName == name.value &&
            toolbar.getAttribute("type") != "menubar" &&
            toolbar.nodeName == 'toolbar') {
          dupeFound = true;
          break;
        }
      }
    }

    if (!dupeFound)
      break;

    message = stringBundle.getFormattedString("enterToolbarDup", [name.value]);
  }

  navToolbox.appendCustomToolbar(name.value, "", true);

  toolboxChanged();

  if (doneButton)
    doneButton.disabled = false;
}

/**
 * Restore the default set of buttons to fixed toolbars,
 * remove all custom toolbars, and rebuild the palette.
 */
restoreDefaultSet = function () {
  // Unwrap the items on the toolbar.
  unwrapToolbarItems();

  docElt = gToolboxDocument.documentElement;

  for(var i = 0; i < gToolboxes.length; i++) {
    var toolbox = gToolboxes[i];

    if (!gStatusbarMode) {
      // Remove all of the customized toolbars.
      var child = toolbox.lastChild;
      while(child) {
        if (child.hasAttribute("customindex")) {

          // Clean up any customizations from the root doc
          handleOptions("remove", child);

          var thisChild = child;
          child = child.previousSibling;
          thisChild.currentSet = "__empty";
          toolbox.removeChild(thisChild);
        } else {
          child = child.previousSibling;
        }
      }
    }

    // Restore the default icon size and mode - meaning Fx main toolbox if none
    var defaultMode = toolbox.hasAttribute("defaultmode") ?
        toolbox.getAttribute("defaultmode") :
        (navToolbox ? navToolbox.getAttribute("defaultmode") : null);
    var defaultIconsSmall = toolbox.hasAttribute("defaulticonsize") ?
        (toolbox.getAttribute("defaulticonsize") == "small" ?
            "small" : "large") :
        (navToolbox ? (navToolbox.getAttribute("defaulticonsize") == "small" ?
            "small" : "large") : "small");

    setAttribute(toolbox, "iconsize", defaultIconsSmall);
    gToolboxDocument.persist(toolbox.id, "iconsize");
    document.getElementById("smallicons").checked = defaultIconsSmall == "small";
    document.getElementById("smallicons").disabled = (defaultMode == "text" ? true : false);
    setAttribute(toolbox, "mode", defaultMode);
    gToolboxDocument.persist(toolbox.id, "mode");
    document.getElementById("modelist").value = defaultMode;

    if (gStatusbarMode) {
      unwrapToolbarItems();
      tt_resetStatusbar();
    }
    else {
      // Restore the defaultset for fixed toolbars.
      var toolbar = toolbox.firstChild;
      while (toolbar) {
        if (isCustomizableToolbar(toolbar)) {
////          if (!toolbar.hasAttribute("customindex")) {
            var defaultSet = toolbar.getAttribute("defaultset");
            if (defaultSet || defaultSet == "")
              toolbar.currentSet = defaultSet;
            //Toolbox has custom binding, otherwise it's some other 3rd party toolbox
            if (typeof toolbar.parentNode.setToolbarIconSize == "function")
              toolbar.parentNode.setToolbarIconSize(toolbar, defaultIconsSmall);
            else
              if (toolbar.getAttribute("currentset") != "__empty") {
                setAttribute(toolbar, "iconsize", defaultIconsSmall);
                gToolboxDocument.persist(toolbar.id, "iconsize");
              }
            // Toolbox has custom binding, otherwise it's some other 3rd party toolbox
            if (typeof toolbar.parentNode.setToolbarMode == "function")
              toolbar.parentNode.setToolbarMode(toolbar, defaultMode);
            else
              if (toolbar.getAttribute("currentset") != "__empty") {
                setAttribute(toolbar, "mode", defaultMode);
                gToolboxDocument.persist(toolbar.id, "mode");
              }
////          }
        }
        // Show nav-bar and bookmarks toolbar if hidden
        if (toolbar.id == "nav-bar" || toolbar.id == "PersonalToolbar")
          toolbar.removeAttribute("collapsed");

        toolbar = toolbar.nextSibling;
      }
    }
  }

  // Now rebuild the palette.
  buildPalette();

  // Now re-wrap the items on the toolbar.
  wrapToolbarItems();

  toolboxChanged("reset");
}

updateIconSize = function (aUseSmallIcons, localDefault) {
  gToolboxIconSize = aUseSmallIcons ? "small" : "large";

  for(var j = 0; j < gToolboxes.length; j++) {
    var toolbox = gToolboxes[j];

    setAttribute(toolbox, "iconsize", gToolboxIconSize);
    gToolboxDocument.persist(toolbox.id, "iconsize");

    for(var i = 0; i < toolbox.childNodes.length; ++i) {
      var toolbar = toolbox.childNodes[i];
      if (isCustomizableToolbar(toolbar)) {
        //Toolbox has custom binding, otherwise it's some other 3rd party toolbox
        if (typeof toolbar.parentNode.setToolbarIconSize == "function")
          toolbar.parentNode.setToolbarIconSize(toolbar, gToolboxIconSize);
        else {
          var toolbarIconSize = (localDefault &&
              toolbar.hasAttribute("defaulticonsize")) ?
                  toolbar.getAttribute("defaulticonsize") : gToolboxIconSize;
          setAttribute(toolbar, "iconsize", toolbarIconSize);
          gToolboxDocument.persist(toolbar.id, "iconsize");
        }
      }
    }
  }
}

updateToolbarMode = function (aModeValue, localDefault) {

  for(var j = 0; j < gToolboxes.length; j++) {
    var toolbox = gToolboxes[j];

    setAttribute(toolbox, "mode", aModeValue);
    gToolboxDocument.persist(toolbox.id, "mode");

    for (var i = 0; i < toolbox.childNodes.length; ++i) {
      var toolbar = toolbox.childNodes[i];
      if (isCustomizableToolbar(toolbar)) {
        // Toolbox has custom binding, otherwise it's some other 3rd party toolbox
        if (typeof toolbar.parentNode.setToolbarMode == "function")
          toolbar.parentNode.setToolbarMode(toolbar, aModeValue);
        else {
          var toolbarMode = (localDefault &&
              toolbar.hasAttribute("defaultmode")) ?
                  toolbar.getAttribute("defaultmode") : aModeValue;
          setAttribute(toolbar, "mode", toolbarMode);
          gToolboxDocument.persist(toolbar.id, "mode");
        }
      }
    }
  }

  var iconSizeCheckbox = document.getElementById("smallicons");
  iconSizeCheckbox.disabled = aModeValue == "text";
}

function isCustomizableToolbar(aElt) {
  return (aElt) && ((aElt.localName == "toolbar" && aElt.getAttribute("customizable")) ||
          aElt.localName == "statusbar");
}

function isToolbarItem(aElt) {
  return aElt.localName == "toolbarbutton" ||
         aElt.localName == "toolbaritem" ||
         aElt.localName == "toolbarseparator" ||
         aElt.localName == "toolbarspring" ||
         aElt.localName == "toolbarspacer" ||
         aElt.localName == "statusbarpanel";
}

///////////////////////////////////////////////////////////////////////////
//// Drag and Drop observers

function onToolbarDragOver(aEvent) {
  var documentId = gToolboxDocument.documentElement.id;
  if (!aEvent.dataTransfer.types.contains("text/toolbarwrapper-id/" + documentId))
    return;

  var toolbar = aEvent.target;
  var dropTarget = aEvent.target;
  while (toolbar && (toolbar.localName != "toolbar" && toolbar.localName != "statusbar")) {
    dropTarget = toolbar;
    toolbar = toolbar.parentNode;
  }

  var previousDragItem = gCurrentDragOverItem;

  // Make sure we are dragging over a customizable toolbar.
  if (!isCustomizableToolbar(toolbar)) {
    gCurrentDragOverItem = null;
    return;
  }

  if (dropTarget.localName == "toolbar" || dropTarget.localName == "statusbar") {
    gCurrentDragOverItem = dropTarget;
  } else {
    gCurrentDragOverItem = null;

    var direction = window.getComputedStyle(dropTarget.parentNode, null).direction;
    var dropTargetCenter = dropTarget.boxObject.x + (dropTarget.boxObject.width / 2);
    var dragAfter;
    if (direction == "ltr")
      dragAfter = aEvent.clientX > dropTargetCenter;
    else
      dragAfter = aEvent.clientX < dropTargetCenter;

    if (dragAfter) {
      gCurrentDragOverItem = dropTarget.nextSibling;
      if (!gCurrentDragOverItem)
        gCurrentDragOverItem = toolbar;
    } else
        gCurrentDragOverItem = dropTarget;
  }

  if (previousDragItem && gCurrentDragOverItem != previousDragItem) {
    setDragActive(previousDragItem, false);
  }

  setDragActive(gCurrentDragOverItem, true);

  aEvent.preventDefault();
  // Otherwise tabs dnd gets the event and won't allow it.
  aEvent.stopPropagation();
}

function onToolbarDrop(aEvent) {
  if (!gCurrentDragOverItem)
    return;

  setDragActive(gCurrentDragOverItem, false);

  var documentId = gToolboxDocument.documentElement.id;
  var draggedItemId = aEvent.dataTransfer.getData("text/toolbarwrapper-id/" + documentId);
  if (gCurrentDragOverItem.id == draggedItemId)
    return;

  var toolbar = aEvent.target;
  while (toolbar.localName != "toolbar" && toolbar.localName != "statusbar")
    toolbar = toolbar.parentNode;

  var draggedPaletteWrapper = document.getElementById("wrapper-"+draggedItemId);
  if (!draggedPaletteWrapper) {
    // The wrapper has been dragged from the toolbar.
    // Get the wrapper from the toolbar document and make sure that
    // it isn't being dropped on itself.
    var wrapper = gToolboxDocument.getElementById("wrapper-"+draggedItemId);
    if (wrapper == gCurrentDragOverItem)
      return;

    // Don't allow static kids (e.g., the menubar) to move.
    if (wrapper.parentNode.firstPermanentChild &&
        wrapper.parentNode.firstPermanentChild.id == wrapper.firstChild.id)
      return;
    if (wrapper.parentNode.lastPermanentChild &&
        wrapper.parentNode.lastPermanentChild.id == wrapper.firstChild.id)
      return;

    // Remove the item from its place in the toolbar.
    var toolbar = wrapper.parentNode;
    if (!gStatusbarMode) {
      toolbar.removeChild(wrapper);
      // Need to keep some width for empty vertical toolbars.
      if (toolbar.parentNode.hasAttribute("vertical") &&
          toolbar.childElementCount == 0)
        toolbar.setAttribute("flex", "1");
    }

    // Determine which toolbar we are dropping on.
    var dropToolbar = null;
    if (gCurrentDragOverItem.localName == "toolbar" ||
        gCurrentDragOverItem.localName == "statusbar")
      dropToolbar = gCurrentDragOverItem;
    else
      dropToolbar = gCurrentDragOverItem.parentNode;

    // Insert the item into the toolbar.
    if (gCurrentDragOverItem != dropToolbar)
      dropToolbar.insertBefore(wrapper, gCurrentDragOverItem);
    else
      dropToolbar.appendChild(wrapper);

    // Remove flex for now non empty vertical toolbar.
    if (dropToolbar.parentNode.hasAttribute("vertical"))
      dropToolbar.removeAttribute("flex");
  }
  else {
    // The item has been dragged from the palette

    if (!gStatusbarMode || isSpecialItem(draggedPaletteWrapper.firstChild)) {
      // Create a new wrapper for the item. We don't know the id yet.
      var wrapper = createWrapper("", gToolboxDocument);

      // Ask the toolbar to clone the item's template, place it inside the
      // wrapper, and insert it in the toolbar.
      var newItem = toolbar.insertItem(draggedItemId,
          gCurrentDragOverItem == toolbar ? null : gCurrentDragOverItem,
              wrapper);

      // Prepare the item and wrapper to look good on the toolbar.
      cleanupItemForToolbar(newItem, wrapper);
      wrapper.id = "wrapper-"+newItem.id;
      wrapper.flex = newItem.flex;

      // Remove flex for now non empty vertical toolbar.
      if (toolbar.parentNode.hasAttribute("vertical"))
        toolbar.removeAttribute("flex");
    } 
    else {
      // Move the item in the statusbar.
      var wrapper = gToolboxDocument.getElementById("wrapper-"+draggedItemId);
      if (gCurrentDragOverItem != gStatusbar)
        gStatusbar.insertBefore(wrapper, gCurrentDragOverItem);
      else
        gStatusbar.appendChild(wrapper);
      wrapper.removeAttribute("itemremoved");
      wrapper.firstChild.removeAttribute("removed");
      wrapper.firstChild.setAttribute("style", "display: -moz-box !important");
    }

    // Remove the wrapper from the palette.
    var currentRow = draggedPaletteWrapper.parentNode;
    if (draggedItemId != "separator" &&
        draggedItemId != "spring" &&
        draggedItemId != "spacer") {
      currentRow.removeChild(draggedPaletteWrapper);

      while (currentRow) {
        // Pull the first child of the next row up
        // into this row.
        var nextRow = currentRow.nextSibling;

        if (!nextRow) {
          var last = currentRow.lastChild;
          var first = currentRow.firstChild;
          if (first == last) {
            // Kill the row.
            currentRow.parentNode.removeChild(currentRow);
            break;
          }

          if (last.localName == "spacer") {
            var flex = last.getAttribute("flex");
            last.setAttribute("flex", ++flex);
            // Reflow doesn't happen for some reason.
            // Trigger it with a hide/show. ICK! -dwh
            last.hidden = true;
            last.hidden = false;
            break;
          }
          else {
            // Make a spacer and give it a flex of 1.
            var spacer = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                                  "spacer");
            spacer.setAttribute("flex", "1");
            currentRow.appendChild(spacer);
          }
          break;
        }

        currentRow.appendChild(nextRow.firstChild);
        currentRow = currentRow.nextSibling;
      }
    }
  }

  gCurrentDragOverItem = null;

  toolboxChanged();
};

function onPaletteDrop(aEvent) {
  var documentId = gToolboxDocument.documentElement.id;
  var itemId = aEvent.dataTransfer.getData("text/toolbarwrapper-id/" + documentId);

  var wrapper = gToolboxDocument.getElementById("wrapper-"+itemId);
  if (wrapper && !wrapper.getAttribute("itemremoved")) {
    // Don't allow static kids (e.g., the menubar) to move.
    if (wrapper.parentNode.firstPermanentChild &&
        wrapper.parentNode.firstPermanentChild.id == wrapper.firstChild.id)
      return;
    if (wrapper.parentNode.lastPermanentChild &&
        wrapper.parentNode.lastPermanentChild.id == wrapper.firstChild.id)
      return;

    var wrapperType = wrapper.getAttribute("type");
    if (wrapperType != "separator" &&
        wrapperType != "spacer" &&
        wrapperType != "spring") {
      appendPaletteItem(document.importNode(wrapper.firstChild, true));
      if (!gStatusbarMode)
        gToolbox.palette.appendChild(wrapper.firstChild);
    }

    // The item was dragged out of the toolbar.
    var toolbar = wrapper.parentNode;
    if (!gStatusbarMode) {
      toolbar.removeChild(wrapper);
      // Need to keep some width for empty vertical toolbars.
      if (toolbar.parentNode.hasAttribute("vertical") &&
          toolbar.childElementCount == 0)
        toolbar.setAttribute("flex", "1");
    }
    else {
      wrapper.firstChild.setAttribute("removed", true);
      wrapper.setAttribute("itemremoved", true);
      wrapper.firstChild.removeAttribute("style");
    }
  }

  toolboxChanged();
}
